/*
    BlockGame behaviour common to J2ME and J2SE.
    This class interacts with the Block table to destroy colour groups
    and animate any blocks appropiate and do some painting activities.
    Copyright (C) 2003 Irene Cummings (i.n.c@ezrs.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package inc.games;

import inc.games.*;
import java.util.*;
import java.lang.*;

//class "carry all", this places all the code common to Java2 and J2ME in one place
//Mainly deals with creating appropiate clipping areas etc...
//This should be an inner class to a JPanel or Canvas type class 
public class BlockGameBase extends TimerTask
{
	protected BlockTable m_BlockTable;
	public final int m_BlockSize;
	public final int m_ColourNo;
	private int m_AnimateRate;
	private Timer m_UpDateTimer=new Timer();

	public ClipArea m_ClipArea;
	private AbstractCanvas m_Canvas;

	public BlockGameBase(AbstractCanvas canvas, int screenWidth, int screenHeight, int blockSize, int numbColours)
	{
		m_BlockSize=blockSize;
		m_ColourNo=numbColours;
		this.Init(canvas, screenWidth,screenHeight,blockSize,numbColours,1);
	}
		
	public BlockGameBase(AbstractCanvas canvas, int screenWidth,int screenHeight,int blockSize,int numbColours,int slowDown)
	{
		m_BlockSize=blockSize;
		m_ColourNo=numbColours;
		this.Init(canvas, screenWidth,screenHeight,blockSize,numbColours,slowDown);
	}

	private void Init(AbstractCanvas canvas, int screenWidth,int screenHeight,int blockSize,int numbColours,int slowDown)
	{
		//System.out.println("-----------BlockGameBase::Init()----------------------------");
		m_BlockTable=new BlockTable(screenWidth/m_BlockSize,screenHeight/m_BlockSize,m_ColourNo,m_BlockSize);
		m_AnimateRate=slowDown;
		m_ClipArea=new ClipArea(screenWidth,screenHeight);
		m_Canvas=canvas;
		m_UpDateTimer.schedule(this,0,m_AnimateRate);
	}

	//respond to event at location x,y. 
	public void Respond2Event(int x, int y)
	{
		boolean sucess=this.DestroyColourArea(x,y);
		if (sucess==true)
		{
			//if not currently updating better start!
			this.wakeup();
		}

	}


	//Find the current score
	public int GetHighScore()
	{
		Vector vtr=m_BlockTable.GetBlocks();
		return (m_BlockTable.m_Size[0])*(m_BlockTable.m_Size[1])-vtr.size();		
	}

	//true if something removed (i.e. more than one block chosen!)
	private boolean DestroyColourArea(int x, int y)
	{
		//System.out.println("DestroyColourArea("+x+","+y+") called at "+System.currentTimeMillis());
		boolean suceeded=false;

		//which row/column does this x,y location represent?
		//(do it here because may have an offset from the screen
		//edge that BlockTable quite rightly knows nothing about!)
		int col=x/m_BlockSize;
		int row=y/m_BlockSize;
		Vector clrGroup=m_BlockTable.GetColourGroup(col,row);
		if (clrGroup!=null && clrGroup.size()>1)
		{
			suceeded=true;
			for (int i=0;i<clrGroup.size();i++)
			{
				Block blk=(Block)clrGroup.elementAt(i);
				m_ClipArea.AddClipArea(blk.x,blk.y,m_BlockSize,m_BlockSize);
			}
			m_BlockTable.BeginAnimation();
			m_Canvas.Repaint();
			m_ClipArea.Reset();
		}
		//System.out.println("DestroyColourArea left at "+System.currentTimeMillis());
		return suceeded;
	}

	//no access specification means for package only
	boolean UpDate()
	{
		//System.out.println("UpDate() called at "+System.currentTimeMillis());
		boolean suceeded=false;

		Vector animBlocks=m_BlockTable.Animate();
		if (animBlocks!=null && animBlocks.size()>0)
		{
			for (int i=0;i<animBlocks.size();i++)
			{
				Block blk=(Block)animBlocks.elementAt(i);
				//ok this sucks, means we need to know blocks dropping down
				//(i.e. not very encapsulated!)
				m_ClipArea.AddClipArea(blk.x,blk.y-1,m_BlockSize,m_BlockSize);
			}

			//System.out.println("repainting...");
			m_Canvas.Repaint();
			m_ClipArea.Reset();
			//System.out.println("...done");
			suceeded=true;
		}
		else
		{
			m_BlockTable.EndAnimation();
			m_BlockTable.MoveColumnsLeft();
			//System.out.println("repainting...");
			m_Canvas.Repaint();
			//System.out.println("...done");
		}
		//System.out.println("UpDate() finished");
		return suceeded;
	}


	public void Paint(AbstractGraphic graphic)
	{
		Vector drawBlocks=m_BlockTable.GetBlocks();
		graphic.SetClip(m_ClipArea);
		this.PaintBackGround(graphic);
		if (drawBlocks!=null)
		{
			for (int i=0;i<drawBlocks.size();i++)
			{
				Block blk=(Block)drawBlocks.elementAt(i);
				graphic.SetColor(blk.colour);
				graphic.FillRect(blk.x,blk.y,m_BlockSize,m_BlockSize);
			}
		}
	}

	protected void PaintBackGround(AbstractGraphic graphic)
	{
		graphic.SetColor(16777215);//white
		graphic.FillRect(m_ClipArea.m_ClipXReset,m_ClipArea.m_ClipYReset,m_ClipArea.m_ClipWReset,m_ClipArea.m_ClipHReset);
	}

	public void run()
	{
		boolean success=this.UpDate();
		if (success==false)
		{
			this.relax();
		}
	}			

	private synchronized void relax()
	{
		try
		{
			//System.out.println("UpdateTask entering wait mode");
			this.wait();
		}
		catch (InterruptedException e)
		{
			//System.out.println("UpdateTask having problems");
		}
	}

	public synchronized void wakeup()
	{
		//System.out.println("UpdateTask being notified");
		this.notify();
	}
}









