/*
    Block Table, Colour Group & Animation Group classes.
    A block being selected will usually require a colour group be
    formed and sometimes an animation group. These classes
    take care of finding these groups and most of the updating of the 
    block table.
    Copyright (C) 2003 Irene Cummings (i.n.c@ezrs.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package inc.games;

import inc.*;
import java.util.*;

public class BlockTable
{
	private Block m_Matrix[][];
	public int m_Size[]=new int[2];
	public int m_BlockSize;

	private ColourGroup m_ColourGroup;
	private AnimGroup m_AnimGroup;

	public BlockTable(int noCols, int noRows,int noColours,int sizeBlocks)
	{
		m_BlockSize=sizeBlocks;
		RadNo myRand=new RadNo(noColours);
		int cSpacing=16777216/noColours;
		m_Matrix=new Block[noCols][noRows];
		for (int col=0;col<noCols;col++)
		{
			for (int row=0;row<noRows;row++)
			{
				//unlike array creation in c++ does not fill
				//in with type....				
				m_Matrix[col][row]=new Block(m_BlockSize*col,m_BlockSize*row,myRand.nextInt()*cSpacing);
			}
		}
		m_Size[0]=noCols;
		m_Size[1]=noRows;
	}

	//find a block relative to that specified....
	public Block GetBlockNearby(Block current,int relRow, int relCol)
	{
		int xRow=current.x/m_BlockSize+relRow;
		int yRow=current.y/m_BlockSize+relCol;
		//System.out.println("What block at ("+xRow+","+yRow+")?");
		return this.GetBlock(xRow,yRow);
	}

	public Block GetBlock(int col, int row)
	{
		Block block=null;
		if (col<m_Size[0] && row<m_Size[1] && col>=0 && row>=0)
		{
			block=m_Matrix[col][row];
		}
		return block;
	}

	//vector of (none-null) blocks in column containing pixNo
	public Vector GetBlockCol4Pix(int pixNo)
	{
		return GetBlockCol(pixNo/m_BlockSize);
	}

	private Vector GetBlockCol(int colNo)
	{
		Vector result=new Vector();
		for (int i=0;i<m_Size[1];i++)
		{
			Block block=m_Matrix[colNo][i];
			if (block!=null)
			{
				result.addElement(block);
			}
		}
		return result;
	}

	public Vector GetBlocks()
	{
		Vector result=new Vector();
		for (int i=0;i<m_Size[1];i++)
		{
			for (int j=0;j<m_Size[0];j++)
			{
				Block block=m_Matrix[j][i];
				if (block!=null)
				{
					result.addElement(block);
				}
			}
		}
		return result;
	}

	public Vector GetColourGroup(int col, int row)
	{
		Vector result=null;
		if (col<m_Size[0] && row<m_Size[1])
		{
			Block beginB=m_Matrix[col][row];
			if (beginB!=null)
			{
				//System.out.println("block select at "+col+","+row+" colour="+beginB.colour);
				m_ColourGroup=new ColourGroup();
				beginB.findNeighbours(m_ColourGroup,this);
				result=m_ColourGroup.GetElements();				
			}
			else
			{
				//System.out.println("null block select at "+col+","+row);
			}
		}
		return result;
	}

	//remove colour group blocks & create animation group so animation can begin....
	public synchronized void BeginAnimation()
	{
		//Remove blocks in m_ColourGroup from matrix
		for (int pix=m_ColourGroup.xPosB;
			pix<=m_ColourGroup.xPosE;pix+=m_BlockSize)
		{
			Vector clrVtr=m_ColourGroup.GetBlockCol4Pix(pix);
			for (int clr=0;clr<clrVtr.size();clr++)
			{
				//ok! Trickyness alert. Determine block index by location
				//and block size!
				int row=((Block)clrVtr.elementAt(clr)).y/m_BlockSize;
				m_Matrix[pix/m_BlockSize][row]=null;	
			}
		}
		if (m_AnimGroup==null)
		{
			m_AnimGroup=new AnimGroup(m_ColourGroup,this);
		}
		else
		{
			//an animation already underway - add to it!
			m_AnimGroup.AddColourGroup(m_ColourGroup,this);
		}
		//no longer need cg so garbage collection can begin
		m_ColourGroup=null;
	}

	public synchronized void EndAnimation()
	{
		m_AnimGroup=null;
	}


	//dont allow matrix to change before its updated 
	public synchronized Vector Animate()
	{
		Vector results=null;
		if (m_AnimGroup!=null)
		{
			synchronized(m_AnimGroup)
			{
				Vector before=m_AnimGroup.Copy();
				boolean updated=m_AnimGroup.UpDate();
				//only return anim group if update was necessary
				if (updated==true)
				{
					results=m_AnimGroup;
					this.UpDateMatrix(before,m_AnimGroup);
				}
			}
		}
		return results;
	}

	//where before is a copy of original animation group, after is the real deal
	//(what the matrix actually is pointing to!)
	private void UpDateMatrix(Vector before, Vector after)
	{
		if (before.size()==after.size())
		{
			for (int i=0;i<before.size();i++)
			{
				//assuming row=blk.x/m_Size (& col has not changed)
				//truncation of integers will make sure this only
				//happens at the appropiate animation intervals
				int beforeRow=((Block)before.elementAt(i)).y/m_BlockSize;
				int afterRow=((Block)after.elementAt(i)).y/m_BlockSize;
				int col=((Block)before.elementAt(i)).x/m_BlockSize;
				if (beforeRow!=afterRow)
				{
					//System.out.println("UpdateMatrix setting "+col+","+beforeRow+","+((Block)before.elementAt(i)).colour+" to null");
					m_Matrix[col][afterRow]=(Block)after.elementAt(i);
					m_Matrix[col][beforeRow]=null;
					//System.out.println("updatematrix actually shifting things!");
				}
				else
				{
					//System.out.println("update matrix left for "+afterRow+","+i);
					//System.out.println("pixel (before,after)=("+((Block)before.elementAt(i)).y+","+((Block)after.elementAt(i)).y+")");
				}
			}
		}
		else
		{
			//System.out.println("UpDateMatrix before/after vectors different sizes!!");
		}
	}

	public void MoveColumnsLeft()
	{
		int newCol=m_Size[0];
		for (int i=0;i<m_Size[0];i++)
		{
			Vector colVtr=this.GetBlockCol(i);
			if (colVtr==null || colVtr.size()==0)
			{
				if (newCol>i)
				{
					newCol=i;
				}
			}
			else
			{
				if (newCol<i)
				{
					this.MoveCol(i,newCol);
					newCol++;
				}
			}
		}
	}

	//move a column to a new position (updating matrix & blocks)
	private void MoveCol(int oldCol, int newCol)
	{
		//System.out.println("moving column "+oldCol+" to col "+newCol);
		for (int i=0;i<m_Size[1];i++)
		{
			m_Matrix[newCol][i]=m_Matrix[oldCol][i];
			m_Matrix[oldCol][i]=null;
			Block blk=m_Matrix[newCol][i];
			if (blk!=null)
			{
				blk.x=newCol*m_BlockSize;
			}
		}
	}
		
}


class ColourGroup extends Hashtable
{
	public int xPosB=-1;
	public int xPosE=-1;

	ColourGroup()
	{
		//noop
	}

	public void AddElement(Block blk)
	{
		Vector vtr=(Vector)this.get(new Integer(blk.x));
		if (vtr==null)
		{
			vtr=new Vector();
			this.put(new Integer(blk.x),vtr);
		}
		vtr.addElement(blk);
		if (blk.x>xPosE)
		{
			xPosE=blk.x;
		}
		if (blk.x<xPosB || xPosB==-1)
		{
			xPosB=blk.x;
		}
	}

	public boolean IsPresent(Block someBlock)
	{
		//System.out.println("IsPresent contains "+this.size()+" elements");
		boolean result=true;
		Vector column=(Vector)this.get(new Integer(someBlock.x));
		if (column==null)
		{
			result=false;
		}
		else
		{
			//System.out.println("column for "+someBlock.x+" contains "+column.size()+" elements");
			result=column.contains(someBlock);
		}
		return result;
	}

	public Vector GetElements()
	{	
		Vector result=new Vector();
		Enumeration enum=this.elements();
		while (enum.hasMoreElements())
		{
			Vector vtr=(Vector)enum.nextElement();
			for (int i=0;i<vtr.size();i++)
			{
				result.addElement(vtr.elementAt(i));
			}
		}
		return result;
	}

	public Vector GetBlockCol4Pix(int xPixelNo)
	{
		return (Vector)this.get(new Integer(xPixelNo));
	}

}


class AnimGroup extends SortedVector
{
	private AnimGroup() {}

	public AnimGroup(ColourGroup cGroup, BlockTable blockT)
	{
		//System.out.println("create update group");
		this.AddColourGroup(cGroup,blockT);
	}

	//this will fail if falling blocks to be added are currently falling!!
	public void AddColourGroup(ColourGroup cGroup, BlockTable blockT)
	{
		//if any in colour group are currently falling remove from list
		//dead blocks dont fall!
		for (int i=0;i<this.size();i++)
		{
			Block element=(Block)this.elementAt(i);
			if (cGroup.IsPresent(element))
			{
				this.removeElementAt(i);
				//System.out.println("cGroup element "+element.Print()+" found in anim group at "+i);
			}
		}

		//System.out.println("Add Colour group begin pixel "+cGroup.xPosB+" to "+cGroup.xPosE);
		for (int pix=cGroup.xPosB;pix<=cGroup.xPosE;pix+=blockT.m_BlockSize)
		{
			Vector blkVtr=blockT.GetBlockCol4Pix(pix);
			Vector clrVtr=cGroup.GetBlockCol4Pix(pix);

			for (int tbl=blkVtr.size()-1;tbl>=0;tbl--)
			{
				Block falling=null;
				for (int clr=0;clr<clrVtr.size();clr++)
				{
					if (((Block)blkVtr.elementAt(tbl)).y<((Block)clrVtr.elementAt(clr)).y)
					{
						//System.out.println("matrix block("+pix/blockT.m_BlockSize+","+tbl+") neededs updating because of colour block("+pix/blockT.m_BlockSize+","+clr+")");
						if (falling==null)
						{
							falling=(Block)blkVtr.elementAt(tbl);
							this.addElement((inc.Comparable)falling,false);
							//System.out.println("AddColourGroup insertion="+inserted);
						}
						falling.IncreaseDrop(blockT.m_BlockSize);
					}
				}
			}
		}
	}

	public boolean UpDate()
	{
		boolean updated=false;
		for (int i=0;i<this.size();i++)
		{
			//if even one element had further to drop consider the
			//update to have suceeded
			if (((Block)this.elementAt(i)).UpDate()==true)
			{
				updated=true;
			}
		}
		//System.out.println("updating anim group and returning "+updated+" from "+this.size()+" elements");
		return updated;
	}	

	//strangly vector does not have a copy constructer or anything similar....
	public AnimGroup Copy()
	{
		AnimGroup cpy=new AnimGroup();
		for (int i=0;i<this.size();i++)
		{
			cpy.addElement(((Block)this.elementAt(i)).Copy());
		}
		return cpy;
	}
}








